"""
    Helper files for internal use - Geometry related.
    Copyright (C) 2022  Andreas Bjelland Berg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from copy import deepcopy
from math import atan2, sqrt, ceil


def find_intersection(line_1_start, line_1_end, line_2_start, line_2_end):
    """Find the intersection of two line segments.

    Method inspired by
    https://stackoverflow.com/questions/563198/how-do-you-detect-where-two-line-segments-intersect#565282
    with credit, quote:
        "Credit: this method is the 2-dimensional specialization of the 3D line
        intersection algorithm from the article 'Intersection of two lines in
        three-space' by Ronald Goldman, published in Graphics Gems, page 304

    Args:
        line_1_start (tuple): Start point of line 1
        line_1_end (tuple): End point of line 1
        line_2_start (tuple): Start point of line 2
        line_2_end (tuple): End point of line 2
    """
    if line_1_start[0] > line_1_end[0]:
        line_1_start, line_1_end = line_1_end, line_1_start
    if line_2_start[0] > line_2_end[0]:
        line_2_start, line_2_end = line_2_end, line_2_start
    
    def cross_product(v, w):
        return v[0]*w[1] - v[1]*w[0]
    
    delta_1 = [line_1_end[0]-line_1_start[0], line_1_end[1]-line_1_start[1]] #r
    delta_2 = [line_2_end[0]-line_2_start[0], line_2_end[1]-line_2_start[1]] #s

    if cross_product(delta_1, delta_2) == 0:
        # Either parallel or collinear -> No intersections we need to consider
        return None
    
    line_difference = [                         # (q - p)
        line_2_start[0] - line_1_start[0],
        line_2_start[1] - line_1_start[1]
    ]
    t = cross_product(line_difference, delta_2) / cross_product(delta_1, delta_2)
    u = cross_product(line_difference, delta_1) / cross_product(delta_1, delta_2)
    # Set to <= to also detect corner points. Set to < to only detect internals
    if 0 < t < 1 and 0 < u < 1:
        return (line_1_start[0] + t * delta_1[0], line_1_start[1] + t * delta_1[1])
    else:
        return None


def get_perpendicular(delta_x, delta_y) -> 'tuple[float, float]':
    """Return the normal vector of delta_x, delta_y pointing 'left'"""
    return -delta_y, delta_x


def get_extruded_points(
            base_point, normal_x, normal_y, cc_size
        ) -> 'tuple[tuple[float, float], tuple[float, float]]':
    # "Normalize" normal_x and normal_y, such that the length of the normal
    # vector = 1
    prev_length = sqrt(normal_x**2 + normal_y**2)
    normal_x = normal_x / prev_length
    normal_y = normal_y / prev_length
    extruded_above = (
        base_point[0] + normal_x * cc_size / 4,
        base_point[1] + normal_y * cc_size / 4
    )
    extruded_below = (
        base_point[0] - normal_x * cc_size / 4,
        base_point[1] - normal_y * cc_size / 4
    )
    return extruded_above, extruded_below

def line_bends_towards_right(start_point, mid_point, end_point) -> bool:
    angle_start_mid = atan2(
        mid_point[1] - start_point[1], mid_point[0] - start_point[0]
    )
    angle_start_end = atan2(
        end_point[1] - start_point[1], end_point[0] - start_point[0]
    )
    return angle_start_mid > angle_start_end


def distance(point_1, point_2) -> float:
    """Compute the distance between two points, using Euclidean distance"""
    return sqrt((point_2[1] - point_1[1])**2 + (point_2[0] - point_1[0])**2)


def calculate_number_of_points(line_length, wanted_segment_size):
    return ceil(line_length / wanted_segment_size) + 1


def get_midpoint(point_1: 'tuple[float, float]',
                 point_2: 'tuple[float, float]') -> 'tuple[float, float]':
    return (
        0.5 * (point_1[0] + point_2[0]),
        0.5 * (point_1[1] + point_2[1])
    )

def split_at_intersections(
        face_constraints: list, cell_constraints: list
        ) -> 'tuple[list, list, dict]':
    """Split a formatted list of constraints at their intersections

    Args:
        constraints (list): Formatted list of contraints. Must have shape:
            [
                [(x11, y11), (x12, y12), ...],
                [(x21, y21), (x22, y22), ...],
                ...
            ]

    Returns:
        list: Formatted face constraints, but with every constraint broken
            into non-intersecting (except possibly at the end) line segments
        list: Formatted cell constraints, but with every constraint broken
            into non-intersecting (except possibly at the end) line segments
        dict: Dictionary of intersection IDs, used to avoid creating multiple
            Gmsh points in each intersection
    """
    # We save intersections in a dict
    # That way, we avoid creating multiple Gmsh point in each intersection
    intersection_IDs = dict()

    # Avoid changing the inputs
    face_const = deepcopy(face_constraints)
    cell_const = deepcopy(cell_constraints)

    def _check_line(line, other_constraints):
        if len(line) < 2:
            # line is a point
            return
        segment_ID = 0
        while segment_ID < len(line) - 1:
            start = line[segment_ID]         # start = (x11, y11)
            end = line[segment_ID+1]
            # Check for intersections in own line
            # We only need to check "forwards", as we have already looked at
            # the combinations behind us
            # We naturally don't check intersections with the current segment
            other_ID = segment_ID + 1
            while other_ID < len(line) - 1:
                other_start = line[other_ID]
                other_end = line[other_ID + 1]

                intersection = find_intersection(start, end, other_start, other_end)
                if intersection is not None:
                    intersection_IDs[intersection] = None
                    # Insert intersection into both segments of the line
                    line.insert(segment_ID + 1, intersection)
                    line.insert(other_ID + 2, intersection)

                    # Get new segment end 
                    end = line[segment_ID + 1]
                other_ID += 1

            # check for intersections in the other constraints
            for other_line in other_constraints:
                other_ID = 0
                while other_ID < len(other_line) - 1:
                    other_start = other_line[other_ID]
                    other_end = other_line[other_ID + 1]

                    intersection = find_intersection(start, end, other_start, other_end)
                    if intersection is not None:
                        intersection_IDs[intersection] = None
                        # Insert intersection into both segments of the line
                        line.insert(segment_ID + 1, intersection)
                        other_line.insert(other_ID + 1, intersection)

                        # Get new segment end 
                        end = line[segment_ID + 1]
                    other_ID += 1
            segment_ID += 1

    # Go through face constraints first
    for line_id, line in enumerate(face_const):
        _check_line(line, face_const[line_id + 1: ] + cell_const)

    # Now check cell constraints against each other
    for line_id, line in enumerate(cell_const):
        _check_line(line, cell_const[line_id + 1: ])
    return face_const, cell_const, intersection_IDs



"""
    Create a 2D Delaunay grid.
    Copyright (C) 2022  Andreas Bjelland Berg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any, Union, Iterable

import gmsh

from gmsh4mrst._arguments import (
    format_constraints, format_shape,
    format_meshing_algorithm, format_recombination_algorithm
)
from gmsh4mrst._geometry import split_at_intersections
from gmsh4mrst._gmsh import (
    create_threshold_field, create_circumference, create_fracture_point,
    create_cell_constraint_point, create_cell_constraint_line
)

def delaunay_grid_2D(
        cell_dimensions: float,
        *,
        shape: list = None,
        face_constraints: Union[
                    'list[list[Iterable]]',
                    'dict[str, float]',
                    'dict[str, Iterable]',
                    'dict[str, dict[str, float]]',
                    'dict[Any, dict[str, Iterable]]'] = None,
        face_constraint_factor: float = 1/3,
        min_FC_threshold_distance: float = 0.05,
        max_FC_threshold_distance: float = 0.2,
        face_intersection_factor: float = None,
        min_intersection_distance: float = None,
        max_intersection_distance: float = None,
        fracture_mesh_sampling: int = 100,
        cell_constraints: Union[
                    'list[list[Iterable]]',
                    'dict[str, float]',
                    'dict[str, Iterable]',
                    'dict[str, dict[str, float]]',
                    'dict[Any, dict[str, Iterable]]'] = None,
        cell_constraint_factor: float = 1/4,
        cell_constraint_parallel_factor: float = None,
        cell_constraint_perpendicular_factor: float = None,
        cell_constraint_point_factor: float = None,
        cell_constraint_refinement_factor: float = None,
        min_CC_threshold_distance: float = 0.05,
        max_CC_threshold_distance: float = 0.2,
        CC_mesh_sampling: int = 100,
        mesh_algorithm: str = "Delaunay",
        recombination_algorithm: str = None,
        savename: str = None,
        run_frontend: bool = False
    ):
    """Create a 2D mesh, with user-supplied face- and cell constraints.

    Face constraints are traced by lines embedded in the resulting surface
    mesh. Cell constraints are traced by a surrounding transfinite mesh, 
    creating a set of quadrilateral cells surrounding the constraint.

    This project was done as part of my Bachelor thesis during spring 2022, in
    order to help provide a secondary backend to the SINTEF-developed MATLAB
    module MRST, and its submodule UPR. For more information about MRST, see
    https://www.sintef.no/projectweb/mrst/. For more information about UPR, see
    https://www.sintef.no/projectweb/mrst/modules/upr/

    Args:
        cell_dimensions (float): Base dimensions of each cell.

        shape (Iterable[float] | Iterable[Iterable[float]] | dict[str, float],
                optional):
            Size or shape of the domain.
            Legal forms:
                Iterable[float]: Shape = (xmax, ymax), and is the size of a 
                    domain starting at (0, 0). Shape must have length 2.
                    Examples:
                        >>> shape = [1, 1]
                        >>> shape = (9001, 14)
                Iterable[Iterable[float]]: Shape is a list of points making up
                    the domain. All points in shape must be ordered either
                    clockwise or counterclockwise, and all points in shape must
                    have length >= 2 (only the first 2 elements are used).
                    Examples:
                        >>> shape = [
                            (0, 0), (0.5, 0.2), (1, 0), (1, 1), (0, 1)
                        ]
                        >>> shape = [
                            (0, 0), (1, 0), (1, 1), (0, 1)
                        ]
                dict[str, float]: Primarily used when complex domains are used
                    in MATLAB. dict must have keys "x" and "y", where the 
                    corresponding list being the x- and y-coordinates of the
                    complex domain.
                    Examples:
                        >>> shape = {
                                "x": [0, 0.5, 1, 1, 0],
                                "y": [0, 0.2, 0, 1, 1]
                            }
                        >>> shape = {"x": [0, 1, 1, 0], "y": [0, 0, 1, 1]}
            If None, shape will be set to the square between (0, 0) and (1, 1).
            Defaults to None.
            NOTE: Any constraints must be wholly within the supplied domain!

        face_constraints (list[Iterable] | dict[str, float] | dict[str, Iterable]
                | dict[str, dict[str, float]] | dict[str, dict[str, Iterable]],
                optional):
            Constraints for the cell faces. Each constraint is the coordinate
            of a surface-trace. The surface is assumed to be linear between 
            the coordinates. The resulting mesh will place sites such that 
            the surfaces are traced by faces of the grid.
            Legal forms:
                list[list[Iterable]]: Primarily used when calling the function
                    from Python. Each element of face constraint is a list of
                    point(s) along the constraint. Note that each element must
                    be a list of points, even if there is only one point!
                    Examples:
                        >>> face_constraints = [
                                [(0.1, 0.1), (0.9, 0.9)],
                                [(0.2, 0.8), (0.5, 0.5), (0.8, 0.2)]
                            ]
                        >>> face_constraints = [
                                [(0.5, 0.5)],
                                [(0.1, 0.1), (0.9, 0.9)]
                            ]
                dict[str, float] | dict[str, list]: Used for only one
                    constraint line, primarily when the function is called
                    from MATLAB. face_constraints must contain the keys "x" and
                    "y", with the corresponding float or list being the x- and
                    y-coordinates of point(s) along the constraint.
                    Examples:
                        >>> face_constraints = {"x": [0.1, 0.9], "y": [0.1, 0.9]}
                        >>> face_constraints = {"x": 0.5, "y": 0.5}
                dict[Any, dict[str, float]] | dict[Any, dict[str, Iterable]]:
                    Used for more than one constraint line, primarily when the
                    function is called from MATLAB. Each element in
                    face_constraints.values() must contain keys "x" and "y",
                    with the corresponding float or list being the x- and y-
                    coordinates of point(s) along the constraint. It does not
                    matter what is used as keys for the top-level dictionary.
                    Examples:
                        >>> face_constraints = {
                            "line": {"x": [0.1, 0.9], "y": [0.1, 0.9]},
                            "point": {"x": 0.5, "y": 0.5},
                            "keys dont matter": {"x": 0.1, "y": 0.9}.
                            2022: {"x": 0.9, "y": 0.1}
                        }
            NOTE: Any constraints must be wholly within the supplied domain!
            If None, face_constraints will be an empty list. Defaults to None.

        face_constraint_factor (float, optional): The size of the cells close
            to the face constraints, as compared to supplied cell_dimensions.
            Cells within min_threshold_distance will have size
            face_constraint_factor * cell_dimensions. Equivalent to FCFactor in
            MRST/UPR/pebiGrid2D. Defaults to 1/3.

        min_FC_threshold_distance (float, optional): Distance from face
            constraints where cell dimensions will start increasing. Defaults
            to 0.05.

        max_FC_threshold_distance (float, optional): Distance from face
            constraints where cell dimensions will be back to their default
            (max) value, i.e. the supplied argument cell_dimensions. Defaults
            to 0.2.

        face_intersection_factor (float, optional): The size of the cells close
            to intersections between face constraints, as compared to supplied
            cell_dimensions. Cells within min_intersection_distance from an 
            intersection will have size face_intersection_factor * cell_dimensions.
            If None, no extra cell shaping will occur around intersections.
            The factor is also used in "breaks" of lines, i.e. if there is a
            sharp "turn" in a line segment. Defaults to None.

        min_intersection_distance (float, optional): Distance from intersections
            where cell dimensions will start increasing. If None, will use
            min_threshold_distance. Defaults to None.

        max_intersection_distance (float, optional): Distance from intersections
            where cell dimensions will be back to their default (max) value,
            i.e. the supplied argument cell_dimensions. If None, will use
            min_threshold_distance. Defaults to None.

        fracture_mesh_sampling (int, optional): The number of points along the
            face constraints should be sampled to calculate the threshold
            distances. Defaults to 100.

        cell_constraints (list[Iterable] | dict[str, float] | dict[str, Iterable]
                | dict[str, dict[str, float]] | dict[str, dict[str, Iterable]],
                optional):
            Constraints for the cell centroids. Each constraint is a line,
            which the method will attempt to place in the center of the returned
            grid. The lines are assumed to be linear between the coordinates.
            If a constraint is only one coordinate, the line is treated as a
            point constraint.
            Legal forms:
                list[list[Iterable]]: Primarily used when calling the function
                    from Python. Each element of cell constraint is a list of
                    point(s) along the constraint. Note that each element must
                    be a list of points, even if there is only one point!
                    Examples:
                        >>> cell_constraints = [
                                [(0.1, 0.1), (0.9, 0.9)],
                                [(0.2, 0.8), (0.5, 0.5), (0.8, 0.2)]
                            ]
                        >>> cell_constraints = [
                                [(0.5, 0.5)],
                                [(0.1, 0.1), (0.9, 0.9)]
                            ]
                dict[str, float] | dict[str, list]: Used for only one
                    constraint line, primarily when the function is called
                    from MATLAB. cell_constraints must contain the keys "x" and
                    "y", with the corresponding float or list being the x- and
                    y-coordinates of point(s) along the constraint.
                    Examples:
                        >>> cell_constraints = {"x": [0.1, 0.9], "y": [0.1, 0.9]}
                        >>> cell_constraints = {"x": 0.5, "y": 0.5}
                dict[Any, dict[str, float]] | dict[Any, dict[str, Iterable]]:
                    Used for more than one constraint line, primarily when the
                    function is called from MATLAB. Each element in
                    cell_constraints.values() must contain keys "x" and "y",
                    with the corresponding float or list being the x- and y-
                    coordinates of point(s) along the constraint. It does not
                    matter what is used as keys for the top-level dictionary.
                    Examples:
                        >>> cell_constraints = {
                            "line": {"x": [0.1, 0.9], "y": [0.1, 0.9]},
                            "point": {"x": 0.5, "y": 0.5},
                            "keys dont matter": {"x": 0.1, "y": 0.9}.
                            2022: {"x": 0.9, "y": 0.1}
                        }
            NOTE: Any constraints must be wholly within the supplied domain!
            If None, cell_constraints will be an empty list. Defaults to None.

        cell_constraint_factor (float, optional): The size used for cells
            around cell constraints, as compared to supplied cell_dimensions.
            Equivalent to CCFactor in MRST/UPR/pebiGrid2D. Defaults to 1/4.

        cell_constraint_parallel_factor (float, optional): The size used
            for the length of the cells within the cell constraints, as
            compared to the supplied cell_dimensions. Overrides
            cell_constraint_factor for lines. If set to None,
            cell_constraint_factor will be used for length. Defaults to None.
        
        cell_constraint_perpendicular_factor (float, optional): The size used
            for the width of the cells within the cell constraints, as
            compared to the supplied cell_dimensions. Overrides
            cell_constraint_factor for lines. If set to None,
            cell_constraint_factor will be used for width. Defaults to None.
            
        cell_constraint_point_factor (float, optional): The size used for cells
            around cell constraint points, as compared to the supplied
            cell_dimensions. Overrides cell_constraint_factor for points. If
            set to None, cell_constraint_factor will be used for points.
            Defaults to None.

        cell_constraint_refinement_factor (float, optional): The cell size in
            the refinement along the cell constraint lines. If set to None,
            cell_dimensions will be used and no refinement will be done.
            Defaults to None.

        min_CC_threshold_distance (float, optional): Distance from cell
            constraints where cell dimensions will start increasing. Defaults
            to 0.05.

        max_CC_threshold_distance (float, optional): Distance from cell
            constraints where cell dimensions will be back to their default
            (max) value, i.e. the supplied argument cell_dimensions. Defaults
            to 0.2.

        CC_mesh_sampling (int, optional): The number of points along the
            cell constraints should be sampled to calculate the threshold
            distances. Defaults to 100.

        mesh_algorithm (str | int, optional): What meshing algorithm should be
            used. Can either be the Gmsh-given ID of the algorithm or a string:
                "MeshAdapt" = 1
                "Automatic" = 2
                "Delaunay" = 5
                "Frontal" = 6
                "BAMG" = 7
                "DelQuad" = 8
            Defaults to "Delaunay".

        recombination_algorithm (str | int, optional). What recombination
            algorithm should be used, and whether recombination should be done.
            Recombination makes Gmsh attempt to create a quadrangle mesh, rather
            than a triangle mesh. Can be either the Gmsh-given ID of the
            algorithm or a string:
                "Simple" = 0
                "Blossom" = 1
                "SimpleFull" = 2
                "BlossomFull" = 3
            If None, no recombination will be done.
            NOTE: Blossom is Gmsh default, but struggles with constraints. It
                may therefore be beneficial to use SimpleQuad, if there are
                constraints.
            NOTE 2: Recombination may lead to weird results for constraints,
                as the recombination is done after constraints have been applied.
                This is especially true for SimpleFull and BlossomFull, who
                automatically perform a coarser mesh, followed by recombination,
                smoothing and subdivision.
                
        savename (str, optional): Name of the saved file. If None, no file will
            be saved. The MATLAB functions assume that the file will be saved
            as "TEMP_Gmsh_MRST.m". Defaults to "TEMP_Gmsh_MRST.m".

        run_frontend (bool, optional): Set to True in order to run the Gmsh
            frontend and show the created mesh. Defaults to False.
    """
    # Handle default values
    if shape is None:
        shape = [(0, 0), (1, 0), (1, 1), (0, 1)]
    if face_constraints is None:
        face_constraints = []
    if min_intersection_distance is None:
        min_intersection_distance = min_FC_threshold_distance
    if max_intersection_distance is None:
        max_intersection_distance = max_FC_threshold_distance
    if cell_constraints is None:
        cell_constraints = []
    if cell_constraint_parallel_factor is None:
        cell_constraint_parallel_factor = cell_constraint_factor
    if cell_constraint_perpendicular_factor is None:
        cell_constraint_perpendicular_factor = cell_constraint_factor
    if cell_constraint_point_factor is None:
        cell_constraint_point_factor = cell_constraint_factor
    if cell_constraint_refinement_factor is None:
        cell_constraint_refinement_factor = cell_dimensions

    # Format Gmsh algorithms
    mesh_algorithm = format_meshing_algorithm(mesh_algorithm)
    recombination_algorithm = format_recombination_algorithm(recombination_algorithm)
    
    # Format all complex arguments
    shape = format_shape(shape)
    face_constraints = format_constraints(face_constraints)
    cell_constraints = format_constraints(cell_constraints)

    # Split all constraints at their intersection
    face_constraints, cell_constraints, intersection_IDs = split_at_intersections(
        face_constraints, cell_constraints
    )
    
    gmsh.initialize()
    gmsh.model.add("gmsh4mrst")

    # Create corners
    corners = [
        gmsh.model.geo.add_point(point[0], point[1], 0) for point in shape
    ]

    # Create circumference
    circumference = create_circumference(corners)

    ##### CREATE FRACTURES (FACE CONSTRAINTS) ######
    fracture_points = []
    fractures = []
    # Create fracture points
    for line in face_constraints:
        # Handle single points their own way
        if len(line) == 1:
            fracture_points.append(
                gmsh.model.geo.add_point(line[0][0], line[0][1], 0)
            )
            continue
        
        # There exists a line with at least 1 segment
        # As line segment [i] starts with the end of line segment [i - 1],
        # we avoid doubling on points by moving fracture_start out of the loop
        fracture_start = create_fracture_point(line[0], intersection_IDs)
        for i in range(1, len(line)):
            fracture_end = create_fracture_point(line[i], intersection_IDs)
            fractures.append(
                gmsh.model.geo.add_line(fracture_start, fracture_end)
            )
            fracture_start = fracture_end

    # Create cell constraints
    cc_loops = []               # Holds line loops - used when making surface
    cc_lines = []               # Holds lines - used when refining mesh
    cc_point_surfaces = []      # Holds surfaces around CC points
    cc_line_surfaces = []       # Holds surfaces around CC lines
    cc_point_size = cell_constraint_point_factor * cell_dimensions  # Cell size around points
    cc_parallel_size = cell_constraint_parallel_factor * cell_dimensions    # Cell size along lines
    cc_perpendicular_size = cell_constraint_perpendicular_factor * cell_dimensions    # Cell size across lines
    for line in cell_constraints:
        if len(line) == 1:
            # line is a single point
            create_cell_constraint_point(
                line[0], cc_point_size, cc_loops, cc_point_surfaces, cc_lines
            )
        
        else:
            # line has at least 1 line segment
            create_cell_constraint_line(
                line, cc_parallel_size, cc_perpendicular_size, cc_loops, cc_line_surfaces, cc_lines
            )

    # Create curve loop of circumference
    circumference_loop = gmsh.model.geo.add_curve_loop(circumference)
    
    # Define surface from circumference
    # We remove the loops created from cell constraints
    surface = gmsh.model.geo.add_plane_surface([circumference_loop, *cc_loops])

    # Synchronize to prepare for embedding fracture face_constraints
    gmsh.model.geo.synchronize()

    # Embed all fractures
    gmsh.model.mesh.embed(1, fractures, 2, surface)
    gmsh.model.mesh.embed(0, fracture_points, 2, surface)

    # Refine mesh around fractures. We want a finer mesh around the fractures,
    # to ensure as close fit of faces as possible. In addition, this helps
    # handle intersecting fractures
    # This is gotten from
    # https://gitlab.onelab.info/gmsh/gmsh/-/blob/master/tutorials/python/t10.py

    fracture_threshold = create_threshold_field(
        point_list=fracture_points,
        curve_list=fractures,
        sampling=fracture_mesh_sampling,
        min_size=face_constraint_factor * cell_dimensions,
        max_size=cell_dimensions,
        min_distance=min_FC_threshold_distance,
        max_distance=max_FC_threshold_distance
    )

    # Add field for intersection points
    intersection_points = list(intersection_IDs.values())
    if face_intersection_factor is not None:
        intersection_min_size = face_intersection_factor * cell_dimensions
    else:
        intersection_min_size = cell_dimensions
    intersection_threshold = create_threshold_field(
        point_list=intersection_points,
        curve_list=None,
        sampling=None,
        min_size=intersection_min_size,
        max_size=cell_dimensions,
        min_distance=min_intersection_distance,
        max_distance=max_intersection_distance
    )

    # Refine mesh around cell constraints as well
    cell_threshold = create_threshold_field(
        None,
        curve_list=cc_lines,
        sampling=CC_mesh_sampling,
        min_size=cell_constraint_refinement_factor * cell_dimensions,
        max_size=cell_dimensions,
        min_distance=min_CC_threshold_distance,
        max_distance=max_CC_threshold_distance,
    )

    # We use the minimum of all fields as our background mesh
    min_field = gmsh.model.mesh.field.add("Min")
    gmsh.model.mesh.field.setNumbers(min_field, "FieldsList", [
        fracture_threshold, intersection_threshold, cell_threshold
    ])
    gmsh.model.mesh.field.setAsBackgroundMesh(min_field)

    # As we define the entire element size in our background mesh, we disable
    # certain on-by-default mesh calculations
    gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)

    # We finally set the meshing algorithm.
    gmsh.option.setNumber("Mesh.Algorithm", mesh_algorithm)

    if recombination_algorithm is not None:
        gmsh.option.setNumber("Mesh.RecombinationAlgorithm", recombination_algorithm)
        gmsh.model.mesh.set_recombine(2, surface)

    # Generate 2D mesh
    gmsh.model.mesh.generate(2)

    # Optionally save
    if savename is not None:
        gmsh.write(savename)
    
    # Optionally run frontend
    if run_frontend:
        gmsh.fltk.run()

    # Always finalize
    gmsh.finalize()



if __name__ == "__main__":
    delaunay_grid_2D(
        0.2, 
        face_constraints=[
            [(0.25, 0.25), (0.4, 0.5), (0.7, 0.7)],
            [(0.8, 0.1), (0.9, 0.2)],
            [(0.2, 0.9), (0.9, 0.1)]
        ],
        cell_constraints=[
            [(0.1, 0.1)],
            [(0.5, 0.8), (0.6, 0.7), (0.7, 0.8), (0.9, 0.6)],
        ],
        shape=[
            (0, 0), (0.5, 0.2), (1, 0), (1, 1), (0, 1)
        ],
        face_constraint_factor = 1/3,
        face_intersection_factor = 1/9,
        mesh_algorithm="DelQuad",
        # recombination_algorithm="simplefull",
        savename=None,
        run_frontend=True)


"""
    Helper files for internal use - Basic assertions.
    Copyright (C) 2022  Andreas Bjelland Berg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

def assert_column_in_dict(face_constraints: dict, column: str):
    if column not in face_constraints.keys():
        raise ValueError(
            f"Dictionary {dict} must contain column `{column}`!"
        )

def assert_columns_have_same_length(face_constraints: dict, col1: str, col2: str):
    if len(face_constraints.get(col1)) != len(face_constraints.get(col2)):
        raise ValueError(
            f"Column `{col1}` and `{col2}` must be the same length! "
            + f"len({col1}): {len(face_constraints.get(col1))}, len({col2}): {len(face_constraints.get(col2))}"
        )



"""
    Create a base mesh, optimized for converting to PEBI later.
    Copyright (C) 2022  Andreas Bjelland Berg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any, Union, Iterable

import gmsh

from gmsh4mrst._arguments import (
    format_constraints, format_shape,
    format_meshing_algorithm, format_recombination_algorithm
)
from gmsh4mrst._geometry import split_at_intersections
from gmsh4mrst._gmsh import (
    create_threshold_field, create_circumference, create_fracture_point,
    create_cell_constraint_point, create_cell_constraint_line
)

def pebi_base_2D(
        cell_dimensions: float,
        *,
        shape: list = None,
        face_constraints: Union[
                    'list[list[Iterable]]',
                    'dict[str, float]',
                    'dict[str, Iterable]',
                    'dict[str, dict[str, float]]',
                    'dict[Any, dict[str, Iterable]]'] = None,
        face_constraint_factor: float = 1/4,
        face_constraint_parallel_factor: float = None,
        face_constraint_perpendicular_factor: float = None,
        face_constraint_point_factor: float = None,
        face_constraint_refinement_factor: float = None,
        min_FC_threshold_distance: float = 0.05,
        max_FC_threshold_distance: float = 0.2,
        face_mesh_sampling: int = 100,
        cell_constraints: Union[
                    'list[list[Iterable]]',
                    'dict[str, float]',
                    'dict[str, Iterable]',
                    'dict[str, dict[str, float]]',
                    'dict[Any, dict[str, Iterable]]'] = None,
        cell_constraint_factor: float = 1/4,
        cell_constraint_parallel_factor: float = None,
        cell_constraint_perpendicular_factor: float = None,
        cell_constraint_point_factor: float = None,
        cell_constraint_refinement_factor: float = None,
        min_CC_threshold_distance: float = 0.05,
        max_CC_threshold_distance: float = 0.2,
        CC_mesh_sampling: int = 100,
        mesh_algorithm: str = "Delaunay",
        recombination_algorithm: str = None,
        savename: str = None,
        run_frontend: bool = False
    ):
    """Create a 2D base for converting to a PEBI grid, with user-supplied face-
    and cell constraints.

    Face constraints are traced a surrounding transfinite mesh, with three
    nodes on each short side. This ensures that the PEBI faces will align with
    the constraint as well as possible.
    Cell constraints are traced by a surrounding transfinite mesh with four
    nodes on each short site. This ensures that the PEBI cells will align with
    the constraint as well as possible. 

    NOTE: This is meant as a base for later conversion to PEBI grid, primarily
    in UPR. As such, face constraints and cell constraints are swapped and
    reinforced, as an attempt at making the conversion as accurate as possible.

    This project was done as part of my Bachelor thesis during spring 2022, in
    order to help provide a secondary backend to the SINTEF-developed MATLAB
    module MRST, and its submodule UPR. For more information about MRST, see
    https://www.sintef.no/projectweb/mrst/. For more information about UPR, see
    https://www.sintef.no/projectweb/mrst/modules/upr/

    Args:
        cell_dimensions (float): Base dimensions of each cell.

        shape (Iterable[float] | Iterable[Iterable[float]] | dict[str, float],
                optional):
            Size or shape of the domain.
            Legal forms:
                Iterable[float]: Shape = (xmax, ymax), and is the size of a 
                    domain starting at (0, 0). Shape must have length 2.
                    Examples:
                        >>> shape = [1, 1]
                        >>> shape = (9001, 14)
                Iterable[Iterable[float]]: Shape is a list of points making up
                    the domain. All points in shape must be ordered either
                    clockwise or counterclockwise, and all points in shape must
                    have length >= 2 (only the first 2 elements are used).
                    Examples:
                        >>> shape = [
                            (0, 0), (0.5, 0.2), (1, 0), (1, 1), (0, 1)
                        ]
                        >>> shape = [
                            (0, 0), (1, 0), (1, 1), (0, 1)
                        ]
                dict[str, float]: Primarily used when complex domains are used
                    in MATLAB. dict must have keys "x" and "y", where the 
                    corresponding list being the x- and y-coordinates of the
                    complex domain.
                    Examples:
                        >>> shape = {
                                "x": [0, 0.5, 1, 1, 0],
                                "y": [0, 0.2, 0, 1, 1]
                            }
                        >>> shape = {"x": [0, 1, 1, 0], "y": [0, 0, 1, 1]}
            If None, shape will be set to the square between (0, 0) and (1, 1).
            Defaults to None.
            NOTE: Any constraints must be wholly within the supplied domain!

        face_constraints (list[Iterable] | dict[str, float] | dict[str, Iterable]
                | dict[str, dict[str, float]] | dict[str, dict[str, Iterable]],
                optional):
            Constraints for the PEBI faces. Each constraint is the coordinate
            of a surface-trace. The surface is assumed to be linear between 
            the coordinates. The resulting mesh will place sites such that 
            the surfaces are traced by faces of the grid when converted to PEBI.
            Legal forms:
                list[list[Iterable]]: Primarily used when calling the function
                    from Python. Each element of face constraint is a list of
                    point(s) along the constraint. Note that each element must
                    be a list of points, even if there is only one point!
                    Examples:
                        >>> face_constraints = [
                                [(0.1, 0.1), (0.9, 0.9)],
                                [(0.2, 0.8), (0.5, 0.5), (0.8, 0.2)]
                            ]
                        >>> face_constraints = [
                                [(0.5, 0.5)],
                                [(0.1, 0.1), (0.9, 0.9)]
                            ]
                dict[str, float] | dict[str, list]: Used for only one
                    constraint line, primarily when the function is called
                    from MATLAB. face_constraints must contain the keys "x" and
                    "y", with the corresponding float or list being the x- and
                    y-coordinates of point(s) along the constraint.
                    Examples:
                        >>> face_constraints = {"x": [0.1, 0.9], "y": [0.1, 0.9]}
                        >>> face_constraints = {"x": 0.5, "y": 0.5}
                dict[Any, dict[str, float]] | dict[Any, dict[str, Iterable]]:
                    Used for more than one constraint line, primarily when the
                    function is called from MATLAB. Each element in
                    face_constraints.values() must contain keys "x" and "y",
                    with the corresponding float or list being the x- and y-
                    coordinates of point(s) along the constraint. It does not
                    matter what is used as keys for the top-level dictionary.
                    Examples:
                        >>> face_constraints = {
                            "line": {"x": [0.1, 0.9], "y": [0.1, 0.9]},
                            "point": {"x": 0.5, "y": 0.5},
                            "keys dont matter": {"x": 0.1, "y": 0.9}.
                            2022: {"x": 0.9, "y": 0.1}
                        }
            NOTE: Any constraints must be wholly within the supplied domain!
            If None, face_constraints will be an empty list. Defaults to None.

        face_constraint_factor (float, optional): The size of the cells close
            to the face constraints, as compared to supplied cell_dimensions.
            Cells within min_threshold_distance will have size
            face_constraint_factor * cell_dimensions. Equivalent to FCFactor in
            MRST/UPR/pebiGrid2D. Defaults to 1/3.

        face_constraint_parallel_factor (float, optional): The size used
            for the length of the cells within the face constraints, as
            compared to the supplied cell_dimensions. Overrides
            face_constraint_factor for lines. If set to None,
            face_constraint_factor will be used for length. Defaults to None.
        
        face_constraint_perpendicular_factor (float, optional): The size used
            for the width of the cells within the face constraints, as
            compared to the supplied face_dimensions. Overrides
            face_constraint_factor for lines. If set to None,
            face_constraint_factor will be used for width. Defaults to None.
            
        face_constraint_point_factor (float, optional): The size used for cells
            around face constraint points, as compared to the supplied
            cell_dimensions. Overrides face_constraint_factor for points. If
            set to None, face_constraint_factor will be used for points.
            Defaults to None.

        face_constraint_refinement_factor (float, optional): The cell size in
            the refinement along the face constraint lines. If set to None,
            cell_dimensions will be used and no refinement will be done.
            Defaults to None.

        min_FC_threshold_distance (float, optional): Distance from face
            constraints where cell dimensions will start increasing. Defaults
            to 0.05.

        max_FC_threshold_distance (float, optional): Distance from face
            constraints where cell dimensions will be back to their default
            (max) value, i.e. the supplied argument cell_dimensions. Defaults
            to 0.2.

        fracture_mesh_sampling (int, optional): The number of points along the
            face constraints should be sampled to calculate the threshold
            distances. Defaults to 100.

        cell_constraints (list[Iterable] | dict[str, float] | dict[str, Iterable]
                | dict[str, dict[str, float]] | dict[str, dict[str, Iterable]],
                optional):
            Constraints for the cell centroids. Each constraint is a line,
            which the method will attempt to place in the center of the returned
            grid when converted to PEBI. The lines are assumed to be linear
            between the coordinates. If a constraint is only one coordinate,
            the line is treated as a point constraint.
            Legal forms:
                list[list[Iterable]]: Primarily used when calling the function
                    from Python. Each element of cell constraint is a list of
                    point(s) along the constraint. Note that each element must
                    be a list of points, even if there is only one point!
                    Examples:
                        >>> cell_constraints = [
                                [(0.1, 0.1), (0.9, 0.9)],
                                [(0.2, 0.8), (0.5, 0.5), (0.8, 0.2)]
                            ]
                        >>> cell_constraints = [
                                [(0.5, 0.5)],
                                [(0.1, 0.1), (0.9, 0.9)]
                            ]
                dict[str, float] | dict[str, list]: Used for only one
                    constraint line, primarily when the function is called
                    from MATLAB. cell_constraints must contain the keys "x" and
                    "y", with the corresponding float or list being the x- and
                    y-coordinates of point(s) along the constraint.
                    Examples:
                        >>> cell_constraints = {"x": [0.1, 0.9], "y": [0.1, 0.9]}
                        >>> cell_constraints = {"x": 0.5, "y": 0.5}
                dict[Any, dict[str, float]] | dict[Any, dict[str, Iterable]]:
                    Used for more than one constraint line, primarily when the
                    function is called from MATLAB. Each element in
                    cell_constraints.values() must contain keys "x" and "y",
                    with the corresponding float or list being the x- and y-
                    coordinates of point(s) along the constraint. It does not
                    matter what is used as keys for the top-level dictionary.
                    Examples:
                        >>> cell_constraints = {
                            "line": {"x": [0.1, 0.9], "y": [0.1, 0.9]},
                            "point": {"x": 0.5, "y": 0.5},
                            "keys dont matter": {"x": 0.1, "y": 0.9}.
                            2022: {"x": 0.9, "y": 0.1}
                        }
            NOTE: Any constraints must be wholly within the supplied domain!
            If None, cell_constraints will be an empty list. Defaults to None.

        cell_constraint_factor (float, optional): The size used for cells
            around cell constraints, as compared to supplied cell_dimensions.
            Equivalent to CCFactor in MRST/UPR/pebiGrid2D. Defaults to 1/4.

        cell_constraint_parallel_factor (float, optional): The size used
            for the length of the cells within the cell constraints, as
            compared to the supplied cell_dimensions. Overrides
            cell_constraint_factor for lines. If set to None,
            cell_constraint_factor will be used for length. Defaults to None.
        
        cell_constraint_perpendicular_factor (float, optional): The size used
            for the width of the cells within the cell constraints, as
            compared to the supplied cell_dimensions. Overrides
            cell_constraint_factor for lines. If set to None,
            cell_constraint_factor will be used for width. Defaults to None.
            
        cell_constraint_point_factor (float, optional): The size used for cells
            around cell constraint points, as compared to the supplied
            cell_dimensions. Overrides cell_constraint_factor for points. If
            set to None, cell_constraint_factor will be used for points.
            Defaults to None.

        cell_constraint_refinement_factor (float, optional): The cell size in
            the refinement along the cell constraint lines. If set to None,
            cell_dimensions will be used and no refinement will be done.
            Defaults to None.

        min_CC_threshold_distance (float, optional): Distance from cell
            constraints where cell dimensions will start increasing. Defaults
            to 0.05.

        max_CC_threshold_distance (float, optional): Distance from cell
            constraints where cell dimensions will be back to their default
            (max) value, i.e. the supplied argument cell_dimensions. Defaults
            to 0.2.

        CC_mesh_sampling (int, optional): The number of points along the
            cell constraints should be sampled to calculate the threshold
            distances. Defaults to 100.

        mesh_algorithm (str | int, optional): What meshing algorithm should be
            used. Can either be the Gmsh-given ID of the algorithm or a string:
                "MeshAdapt" = 1
                "Automatic" = 2
                "Delaunay" = 5
                "Frontal" = 6
                "BAMG" = 7
                "DelQuad" = 8
            Defaults to "Delaunay".

        recombination_algorithm (str | int, optional). What recombination
            algorithm should be used, and whether recombination should be done.
            Recombination makes Gmsh attempt to create a quadrangle mesh, rather
            than a triangle mesh. Can be either the Gmsh-given ID of the
            algorithm or a string:
                "Simple" = 0
                "Blossom" = 1
                "SimpleFull" = 2
                "BlossomFull" = 3
            If None, no recombination will be done.
            NOTE: Blossom is Gmsh default, but struggles with constraints. It
                may therefore be beneficial to use SimpleQuad, if there are
                constraints.
            NOTE 2: Recombination may lead to weird results for constraints,
                as the recombination is done after constraints have been applied.
                This is especially true for SimpleFull and BlossomFull, who
                automatically perform a coarser mesh, followed by recombination,
                smoothing and subdivision.
                
        savename (str, optional): Name of the saved file. If None, no file will
            be saved. The MATLAB functions assume that the file will be saved
            as "TEMP_Gmsh_MRST.m". Defaults to "TEMP_Gmsh_MRST.m".

        run_frontend (bool, optional): Set to True in order to run the Gmsh
            frontend and show the created mesh. Defaults to False.
    """
    # Handle default values
    if shape is None:
        shape = [(0, 0), (1, 0), (1, 1), (0, 1)]
    if face_constraints is None:
        face_constraints = []
    if face_constraint_parallel_factor is None:
        face_constraint_parallel_factor = face_constraint_factor
    if face_constraint_perpendicular_factor is None:
        face_constraint_perpendicular_factor = face_constraint_factor
    if face_constraint_point_factor is None:
        face_constraint_point_factor = face_constraint_factor
    if face_constraint_refinement_factor is None:
        face_constraint_refinement_factor = cell_dimensions
    if cell_constraints is None:
        cell_constraints = []
    if cell_constraint_parallel_factor is None:
        cell_constraint_parallel_factor = cell_constraint_factor
    if cell_constraint_perpendicular_factor is None:
        cell_constraint_perpendicular_factor = cell_constraint_factor
    if cell_constraint_point_factor is None:
        cell_constraint_point_factor = cell_constraint_factor
    if cell_constraint_refinement_factor is None:
        cell_constraint_refinement_factor = cell_dimensions

    # Format Gmsh algorithms
    mesh_algorithm = format_meshing_algorithm(mesh_algorithm)
    recombination_algorithm = format_recombination_algorithm(recombination_algorithm)
    
    # Format all complex arguments
    shape = format_shape(shape)
    face_constraints = format_constraints(face_constraints)
    cell_constraints = format_constraints(cell_constraints)

    # Split all constraints at their intersection
    face_constraints, cell_constraints, intersection_IDs = split_at_intersections(
        face_constraints, cell_constraints
    )
    
    gmsh.initialize()
    gmsh.model.add("gmsh4mrst")

    # Create corners
    corners = [
        gmsh.model.geo.add_point(point[0], point[1], 0) for point in shape
    ]

    # Create circumference
    circumference = create_circumference(corners)

    ##### CREATE FRACTURES (FACE CONSTRAINTS) ######
    face_loops = []               # Holds line loops - used when making surface
    face_lines = []               # Holds lines - used when refining mesh
    face_point_surfaces = []      # Holds surfaces around CC points
    face_line_surfaces = []       # Holds surfaces around CC lines
    face_point_size = face_constraint_point_factor * cell_dimensions                    # Cell size around points
    face_parallel_size = face_constraint_parallel_factor * cell_dimensions              # Cell size along lines
    face_perpendicular_size = face_constraint_perpendicular_factor * cell_dimensions    # Cell size across lines
    for line in face_constraints:
        if len(line) == 1:
            # line is a single point
            create_cell_constraint_point(
                line[0], face_point_size, face_loops, face_point_surfaces, face_lines,
                nodes=3
            )
        
        else:
            # line has at least 1 line segment
            create_cell_constraint_line(
                line, face_parallel_size, face_perpendicular_size, face_loops,
                face_line_surfaces, face_lines, perp_nodes=3
            )

    # Create cell constraints
    cc_loops = []               # Holds line loops - used when making surface
    cc_lines = []               # Holds lines - used when refining mesh
    cc_point_surfaces = []      # Holds surfaces around CC points
    cc_line_surfaces = []       # Holds surfaces around CC lines
    cc_point_size = cell_constraint_point_factor * cell_dimensions  # Cell size around points
    cc_parallel_size = cell_constraint_parallel_factor * cell_dimensions    # Cell size along lines
    cc_perpendicular_size = cell_constraint_perpendicular_factor * cell_dimensions    # Cell size across lines
    for line in cell_constraints:
        if len(line) == 1:
            # line is a single point
            create_cell_constraint_point(
                line[0], cc_point_size, cc_loops, cc_point_surfaces, cc_lines,
                nodes=4
            )
        
        else:
            # line has at least 1 line segment
            create_cell_constraint_line(
                line, cc_parallel_size, cc_perpendicular_size, cc_loops,
                cc_line_surfaces, cc_lines, perp_nodes=4
            )

    # Create curve loop of circumference
    circumference_loop = gmsh.model.geo.add_curve_loop(circumference)
    
    # Define surface from circumference
    # We remove the loops created from the constraints
    surface = gmsh.model.geo.add_plane_surface([circumference_loop, *face_loops, *cc_loops])

    # Synchronize to prepare for embedding fracture face_constraints
    gmsh.model.geo.synchronize()

    # Refine mesh around fractures. We want a finer mesh around the fractures,
    # to ensure as close fit of faces as possible. In addition, this helps
    # handle intersecting fractures
    # This is gotten from
    # https://gitlab.onelab.info/gmsh/gmsh/-/blob/master/tutorials/python/t10.py

    face_threshold = create_threshold_field(
        point_list=None,
        curve_list=face_lines,
        sampling=face_mesh_sampling,
        min_size=face_constraint_refinement_factor * cell_dimensions,
        max_size=cell_dimensions,
        min_distance=min_FC_threshold_distance,
        max_distance=max_FC_threshold_distance
    )


    # Refine mesh around cell constraints as well
    cell_threshold = create_threshold_field(
        None,
        curve_list=cc_lines,
        sampling=CC_mesh_sampling,
        min_size=cell_constraint_refinement_factor * cell_dimensions,
        max_size=cell_dimensions,
        min_distance=min_CC_threshold_distance,
        max_distance=max_CC_threshold_distance,
    )

    # We use the minimum of all fields as our background mesh
    min_field = gmsh.model.mesh.field.add("Min")
    gmsh.model.mesh.field.setNumbers(min_field, "FieldsList", [
        face_threshold, cell_threshold
    ])
    gmsh.model.mesh.field.setAsBackgroundMesh(min_field)

    # As we define the entire element size in our background mesh, we disable
    # certain on-by-default mesh calculations
    gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)

    # We finally set the meshing algorithm.
    gmsh.option.setNumber("Mesh.Algorithm", mesh_algorithm)

    if recombination_algorithm is not None:
        gmsh.option.setNumber("Mesh.RecombinationAlgorithm", recombination_algorithm)
        gmsh.model.mesh.set_recombine(2, surface)

    # Generate 2D mesh
    gmsh.model.mesh.generate(2)

    # Optionally save
    if savename is not None:
        gmsh.write(savename)
    
    # Optionally run frontend
    if run_frontend:
        gmsh.fltk.run()

    # Always finalize
    gmsh.finalize()



if __name__ == "__main__":
    pebi_base_2D(
        0.2, 
        face_constraints=[
            [(0.25, 0.25), (0.75, 0.75)],
        ],
        cell_constraints=[
            [(0.1, 0.9), (0.9, 0.9)],
        ],
        shape=[
            (0, 0), (0.5, 0.2), (1, 0), (1, 1), (0, 1)
        ],
        mesh_algorithm="DelQuad",
        # recombination_algorithm="simplefull",
        savename=None,
        run_frontend=True)


"""
    Create a 2D background mesh, without any embedding.
    Copyright (C) 2022  Andreas Bjelland Berg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Any, Union, Iterable

import gmsh

from gmsh4mrst._arguments import (
    format_constraints, format_shape,
    format_meshing_algorithm, format_recombination_algorithm
)
from gmsh4mrst._geometry import split_at_intersections
from gmsh4mrst._gmsh import (
    create_threshold_field, create_circumference, create_fracture_point,
    create_cell_constraint_point, create_cell_constraint_line
)

def background_grid_2D(
        cell_dimensions: float,
        *,
        shape: list = None,
        face_constraints: Union[
                    'list[list[Iterable]]',
                    'dict[str, float]',
                    'dict[str, Iterable]',
                    'dict[str, dict[str, float]]',
                    'dict[Any, dict[str, Iterable]]'] = None,
        face_constraint_factor: float = 1/3,
        min_FC_threshold_distance: float = 0.05,
        max_FC_threshold_distance: float = 0.2,
        face_intersection_factor: float = None,
        min_intersection_distance: float = None,
        max_intersection_distance: float = None,
        fracture_mesh_sampling: int = 100,
        cell_constraints: Union[
                    'list[list[Iterable]]',
                    'dict[str, float]',
                    'dict[str, Iterable]',
                    'dict[str, dict[str, float]]',
                    'dict[Any, dict[str, Iterable]]'] = None,
        cell_constraint_factor: float = 1/4,
        cell_constraint_line_factor: float = None,
        cell_constraint_point_factor: float = None,
        min_CC_threshold_distance: float = 0.05,
        max_CC_threshold_distance: float = 0.2,
        CC_mesh_sampling: int = 100,
        mesh_algorithm: str = "Delaunay",
        recombination_algorithm: str = None,
        savename: str = None,
        run_frontend: bool = False
    ):
    """Create a 2D mesh, with user-supplied face- and cell constraints.

    Unlike delaunay_grid_2D, no constraints are actually embedded in the final
    result. Instead, only grid refinement is done. This is done due to the
    conversion from Delaunay triangulations to PEBI grids losing information
    about cell faces, leaving constraints messed up. By only refining the grid,
    we can instead handle constraints in MATLAB.

    This project was done as part of my Bachelor thesis during spring 2022, in
    order to help provide a secondary backend to the SINTEF-developed MATLAB
    module MRST, and its submodule UPR. For more information about MRST, see
    https://www.sintef.no/projectweb/mrst/. For more information about UPR, see
    https://www.sintef.no/projectweb/mrst/modules/upr/

    Args:
        cell_dimensions (float): Base dimensions of each cell.

        shape (Iterable[float] | Iterable[Iterable[float]] | dict[str, float],
                optional):
            Size or shape of the domain.
            Legal forms:
                Iterable[float]: Shape = (xmax, ymax), and is the size of a 
                    domain starting at (0, 0). Shape must have length 2.
                    Examples:
                        >>> shape = [1, 1]
                        >>> shape = (9001, 14)
                Iterable[Iterable[float]]: Shape is a list of points making up
                    the domain. All points in shape must be ordered either
                    clockwise or counterclockwise, and all points in shape must
                    have length >= 2 (only the first 2 elements are used).
                    Examples:
                        >>> shape = [
                            (0, 0), (0.5, 0.2), (1, 0), (1, 1), (0, 1)
                        ]
                        >>> shape = [
                            (0, 0), (1, 0), (1, 1), (0, 1)
                        ]
                dict[str, float]: Primarily used when complex domains are used
                    in MATLAB. dict must have keys "x" and "y", where the 
                    corresponding list being the x- and y-coordinates of the
                    complex domain.
                    Examples:
                        >>> shape = {
                                "x": [0, 0.5, 1, 1, 0],
                                "y": [0, 0.2, 0, 1, 1]
                            }
                        >>> shape = {"x": [0, 1, 1, 0], "y": [0, 0, 1, 1]}
            If None, shape will be set to the square between (0, 0) and (1, 1).
            Defaults to None.
            NOTE: Any constraints must be wholly within the supplied domain!

        face_constraints (list[Iterable] | dict[str, float] | dict[str, Iterable]
                | dict[str, dict[str, float]] | dict[str, dict[str, Iterable]],
                optional):
            Constraints for the cell faces. Each constraint is the coordinate
            of a surface-trace. The surface is assumed to be linear between 
            the coordinates. The resulting mesh will place sites such that 
            the surfaces are traced by faces of the grid.
            Legal forms:
                list[list[Iterable]]: Primarily used when calling the function
                    from Python. Each element of face constraint is a list of
                    point(s) along the constraint. Note that each element must
                    be a list of points, even if there is only one point!
                    Examples:
                        >>> face_constraints = [
                                [(0.1, 0.1), (0.9, 0.9)],
                                [(0.2, 0.8), (0.5, 0.5), (0.8, 0.2)]
                            ]
                        >>> face_constraints = [
                                [(0.5, 0.5)],
                                [(0.1, 0.1), (0.9, 0.9)]
                            ]
                dict[str, float] | dict[str, list]: Used for only one
                    constraint line, primarily when the function is called
                    from MATLAB. face_constraints must contain the keys "x" and
                    "y", with the corresponding float or list being the x- and
                    y-coordinates of point(s) along the constraint.
                    Examples:
                        >>> face_constraints = {"x": [0.1, 0.9], "y": [0.1, 0.9]}
                        >>> face_constraints = {"x": 0.5, "y": 0.5}
                dict[Any, dict[str, float]] | dict[Any, dict[str, Iterable]]:
                    Used for more than one constraint line, primarily when the
                    function is called from MATLAB. Each element in
                    face_constraints.values() must contain keys "x" and "y",
                    with the corresponding float or list being the x- and y-
                    coordinates of point(s) along the constraint. It does not
                    matter what is used as keys for the top-level dictionary.
                    Examples:
                        >>> face_constraints = {
                            "line": {"x": [0.1, 0.9], "y": [0.1, 0.9]},
                            "point": {"x": 0.5, "y": 0.5},
                            "keys dont matter": {"x": 0.1, "y": 0.9}.
                            2022: {"x": 0.9, "y": 0.1}
                        }
            NOTE: Any constraints must be wholly within the supplied domain!
            If None, face_constraints will be an empty list. Defaults to None.

        face_constraint_factor (float, optional): The size of the cells close
            to the face constraints, as compared to supplied cell_dimensions.
            Cells within min_threshold_distance will have size
            face_constraint_factor * cell_dimensions. Equivalent to FCFactor in
            MRST/UPR/pebiGrid2D. Defaults to 1/3.

        min_FC_threshold_distance (float, optional): Distance from face
            constraints where cell dimensions will start increasing. Defaults
            to 0.05.

        max_FC_threshold_distance (float, optional): Distance from face
            constraints where cell dimensions will be back to their default
            (max) value, i.e. the supplied argument cell_dimensions. Defaults
            to 0.2.

        face_intersection_factor (float, optional): The size of the cells close
            to intersections between face constraints, as compared to supplied
            cell_dimensions. Cells within min_intersection_distance from an 
            intersection will have size face_intersection_factor * cell_dimensions.
            If None, no extra cell shaping will occur around intersections.
            The factor is also used in "breaks" of lines, i.e. if there is a
            sharp "turn" in a line segment. Defaults to None.

        min_intersection_distance (float, optional): Distance from intersections
            where cell dimensions will start increasing. If None, will use
            min_threshold_distance. Defaults to None.

        max_intersection_distance (float, optional): Distance from intersections
            where cell dimensions will be back to their default (max) value,
            i.e. the supplied argument cell_dimensions. If None, will use
            min_threshold_distance. Defaults to None.

        fracture_mesh_sampling (int, optional): The number of points along the
            face constraints should be sampled to calculate the threshold
            distances. Defaults to 100.

        cell_constraints (list[Iterable] | dict[str, float] | dict[str, Iterable]
                | dict[str, dict[str, float]] | dict[str, dict[str, Iterable]],
                optional):
            Constraints for the cell centroids. Each constraint is a line,
            which the method will attempt to place in the center of the returned
            grid. The lines are assumed to be linear between the coordinates.
            If a constraint is only one coordinate, the line is treated as a
            point constraint.
            Legal forms:
                list[list[Iterable]]: Primarily used when calling the function
                    from Python. Each element of cell constraint is a list of
                    point(s) along the constraint. Note that each element must
                    be a list of points, even if there is only one point!
                    Examples:
                        >>> cell_constraints = [
                                [(0.1, 0.1), (0.9, 0.9)],
                                [(0.2, 0.8), (0.5, 0.5), (0.8, 0.2)]
                            ]
                        >>> cell_constraints = [
                                [(0.5, 0.5)],
                                [(0.1, 0.1), (0.9, 0.9)]
                            ]
                dict[str, float] | dict[str, list]: Used for only one
                    constraint line, primarily when the function is called
                    from MATLAB. cell_constraints must contain the keys "x" and
                    "y", with the corresponding float or list being the x- and
                    y-coordinates of point(s) along the constraint.
                    Examples:
                        >>> cell_constraints = {"x": [0.1, 0.9], "y": [0.1, 0.9]}
                        >>> cell_constraints = {"x": 0.5, "y": 0.5}
                dict[Any, dict[str, float]] | dict[Any, dict[str, Iterable]]:
                    Used for more than one constraint line, primarily when the
                    function is called from MATLAB. Each element in
                    cell_constraints.values() must contain keys "x" and "y",
                    with the corresponding float or list being the x- and y-
                    coordinates of point(s) along the constraint. It does not
                    matter what is used as keys for the top-level dictionary.
                    Examples:
                        >>> cell_constraints = {
                            "line": {"x": [0.1, 0.9], "y": [0.1, 0.9]},
                            "point": {"x": 0.5, "y": 0.5},
                            "keys dont matter": {"x": 0.1, "y": 0.9}.
                            2022: {"x": 0.9, "y": 0.1}
                        }
            NOTE: Any constraints must be wholly within the supplied domain!
            If None, cell_constraints will be an empty list. Defaults to None.

        cell_constraint_factor (float, optional): The size used for cells
            around cell constraints, as compared to supplied cell_dimensions.
            Equivalent to CCFactor in MRST/UPR/pebiGrid2D. Defaults to 1/4.

        cell_constraint_line_factor (float, optional): The size used for cells
            around the cell constraint lines, as compared to the supplied
            cell_dimensions. Overrides cell_constraint_factor for lines. If set
            to None, cell_constraint_factor will be used for lines. Defaults
            to None.
        
        cell_constraint_point_factor (float, optional): The size used for cells
            around cell constraint points, as compared to the supplied
            cell_dimensions. Overrides cell_constraint_factor for points. If
            set to None, cell_constraint_factor will be used for points.
            Defaults to None.

        min_CC_threshold_distance (float, optional): Distance from cell
            constraints where cell dimensions will start increasing. Defaults
            to 0.05.

        max_CC_threshold_distance (float, optional): Distance from cell
            constraints where cell dimensions will be back to their default
            (max) value, i.e. the supplied argument cell_dimensions. Defaults
            to 0.2.

        CC_mesh_sampling (int, optional): The number of points along the
            cell constraints should be sampled to calculate the threshold
            distances. Defaults to 100.

        mesh_algorithm (str | int, optional): What meshing algorithm should be
            used. Can either be the Gmsh-given ID of the algorithm or a string:
                "MeshAdapt" = 1
                "Automatic" = 2
                "Delaunay" = 5
                "Frontal" = 6
                "BAMG" = 7
                "DelQuad" = 8
            Defaults to "Delaunay".

        recombination_algorithm (str | int, optional). What recombination
            algorithm should be used, and whether recombination should be done.
            Recombination makes Gmsh attempt to create a quadrangle mesh, rather
            than a triangle mesh. Can be either the Gmsh-given ID of the
            algorithm or a string:
                "Simple" = 0
                "Blossom" = 1
                "SimpleFull" = 2
                "BlossomFull" = 3
            If None, no recombination will be done.
            NOTE: Blossom is Gmsh default, but struggles with constraints. It
                may therefore be beneficial to use SimpleQuad, if there are
                constraints.
            NOTE 2: Recombination may lead to weird results for constraints,
                as the recombination is done after constraints have been applied.
                This is especially true for SimpleFull and BlossomFull, who
                automatically perform a coarser mesh, followed by recombination,
                smoothing and subdivision.
                
        savename (str, optional): Name of the saved file. If None, no file will
            be saved. Defaults to None.

        run_frontend (bool, optional): Set to True in order to run the Gmsh
            frontend and show the created mesh. Defaults to False.
    """
    # Handle default values
    if shape is None:
        shape = [(0, 0), (1, 0), (1, 1), (0, 1)]
    if face_constraints is None:
        face_constraints = []
    if min_intersection_distance is None:
        min_intersection_distance = min_FC_threshold_distance
    if max_intersection_distance is None:
        max_intersection_distance = max_FC_threshold_distance
    if cell_constraints is None:
        cell_constraints = []
    if cell_constraint_line_factor is None:
        cell_constraint_line_factor = cell_constraint_factor
    if cell_constraint_point_factor is None:
        cell_constraint_point_factor = cell_constraint_factor

    # Format Gmsh algorithms
    mesh_algorithm = format_meshing_algorithm(mesh_algorithm)
    recombination_algorithm = format_recombination_algorithm(recombination_algorithm)
    
    # Format all complex arguments
    shape = format_shape(shape)
    face_constraints = format_constraints(face_constraints)
    cell_constraints = format_constraints(cell_constraints)

    # Split all constraints at their intersection
    face_constraints, cell_constraints, intersection_IDs = split_at_intersections(
        face_constraints, cell_constraints
    )
    
    gmsh.initialize()
    gmsh.model.add("gmsh4mrst")

    # Create corners
    corners = [
        gmsh.model.geo.add_point(point[0], point[1], 0) for point in shape
    ]

    # Create circumference
    circumference = create_circumference(corners)

    ##### CREATE FRACTURES (FACE CONSTRAINTS) ######
    fracture_points = []
    fractures = []
    # Create fracture points
    for line in face_constraints:
        # Handle single points their own way
        if len(line) == 1:
            fracture_points.append(
                gmsh.model.geo.add_point(line[0][0], line[0][1], 0)
            )
            continue
        
        # There exists a line with at least 1 segment
        # As line segment [i] starts with the end of line segment [i - 1],
        # we avoid doubling on points by moving fracture_start out of the loop
        fracture_start = create_fracture_point(line[0], intersection_IDs)
        for i in range(1, len(line)):
            fracture_end = create_fracture_point(line[i], intersection_IDs)
            fractures.append(
                gmsh.model.geo.add_line(fracture_start, fracture_end)
            )
            fracture_start = fracture_end

    # Create cell constraints\
    cc_lines = []               # Holds lines - used when refining mesh
    cc_points = []              # Holds single points - used when refining mesh
    cc_point_size = cell_constraint_point_factor * cell_dimensions  # Cell size around points
    cc_line_size = cell_constraint_line_factor * cell_dimensions    # Cell size around lines
    for line in cell_constraints:
        if len(line) == 1:
            # line is a single point
            cc_points.append(
                gmsh.model.geo.add_point(line[0][0], line[0][1], 0)
            )
        
        else:
            # line has at least 1 line segment
            line_start = create_fracture_point(line[0], intersection_IDs)
            for i in range(1, len(line)):
                line_end = create_fracture_point(line[i], intersection_IDs)
                cc_lines.append(
                    gmsh.model.geo.add_line(line_start, line_end)
                )
                line_start = line_end

    # Create curve loop of circumference
    circumference_loop = gmsh.model.geo.add_curve_loop(circumference)
    
    # Define surface from circumference
    surface = gmsh.model.geo.add_plane_surface([circumference_loop])

    # Synchronize to prepare for embedding fracture face_constraints
    gmsh.model.geo.synchronize()

    # Refine mesh around fractures. We want a finer mesh around the fractures,
    # to ensure as close fit of faces as possible. In addition, this helps
    # handle intersecting fractures
    # This is gotten from
    # https://gitlab.onelab.info/gmsh/gmsh/-/blob/master/tutorials/python/t10.py
    fracture_threshold = create_threshold_field(
        point_list=fracture_points,
        curve_list=fractures,
        sampling=fracture_mesh_sampling,
        min_size=face_constraint_factor * cell_dimensions,
        max_size=cell_dimensions,
        min_distance=min_FC_threshold_distance,
        max_distance=max_FC_threshold_distance
    )

    # Add field for intersection points
    intersection_points = list(intersection_IDs.values())
    if face_intersection_factor is not None:
        intersection_min_size = face_intersection_factor * cell_dimensions
    else:
        intersection_min_size = cell_dimensions
    intersection_threshold = create_threshold_field(
        point_list=intersection_points,
        curve_list=None,
        sampling=None,
        min_size=intersection_min_size,
        max_size=cell_dimensions,
        min_distance=min_intersection_distance,
        max_distance=max_intersection_distance
    )

    # Refine mesh around cell constraint lines
    cell_line_threshold = create_threshold_field(
        None,
        curve_list=cc_lines,
        sampling=CC_mesh_sampling,
        min_size=cc_line_size,
        max_size=cell_dimensions,
        min_distance=min_CC_threshold_distance,
        max_distance=max_CC_threshold_distance,
    )

    # Refine mesh around cell constraint points
    cell_point_threshold = create_threshold_field(
        cc_points,
        curve_list=None,
        sampling=CC_mesh_sampling,
        min_size=cc_point_size,
        max_size=cell_dimensions,
        min_distance=min_CC_threshold_distance,
        max_distance=max_CC_threshold_distance,
    )

    # We use the minimum of all fields as our background mesh
    min_field = gmsh.model.mesh.field.add("Min")
    gmsh.model.mesh.field.setNumbers(min_field, "FieldsList", [
        fracture_threshold, intersection_threshold,
        cell_line_threshold, cell_point_threshold
    ])
    gmsh.model.mesh.field.setAsBackgroundMesh(min_field)

    # As we define the entire element size in our background mesh, we disable
    # certain on-by-default mesh calculations
    gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)

    # We finally set the meshing algorithm.
    gmsh.option.setNumber("Mesh.Algorithm", mesh_algorithm)

    if recombination_algorithm is not None:
        gmsh.option.setNumber("Mesh.RecombinationAlgorithm", recombination_algorithm)
        gmsh.model.mesh.set_recombine(2, surface)

    # Generate 2D mesh
    gmsh.model.mesh.generate(2)

    # Optionally save
    if savename is not None:
        gmsh.write(savename)
    
    # Optionally run frontend
    if run_frontend:
        gmsh.fltk.run()

    # Always finalize
    gmsh.finalize()



if __name__ == "__main__":
    background_grid_2D(
        0.2, 
        face_constraints=[
            [(0.25, 0.25), (0.4, 0.5), (0.7, 0.7)],
            [(0.8, 0.1), (0.9, 0.2)],
            [(0.2, 0.9), (0.9, 0.1)]
        ],
        cell_constraints=[
            [(0.1, 0.1)],
            [(0.5, 0.8), (0.6, 0.7), (0.7, 0.8), (0.9, 0.6)],
        ],
        shape=[
            (0, 0), (0.5, 0.2), (1, 0), (1, 1), (0, 1)
        ],
        mesh_algorithm="DelQuad",
        # recombination_algorithm="simplefull",
        savename=None,
        run_frontend=True)


"""
    Automatically create Gmsh meshes

    These can be used directly in a Python program, but are mainly designed to
    work with the MATLAB module MRST.  
    
    FUNCTIONS:
        background_grid_2D
        delaunay_grid_2D
        pebi_base_2D

    LICENSE:
    Copyright (C) 2022  Andreas Bjelland Berg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .background_grid_2D import background_grid_2D
from .delaunay_grid_2D import delaunay_grid_2D
from .pebi_base_2D import pebi_base_2D


"""
    Helper files for internal use - Gmsh related.
    Copyright (C) 2022  Andreas Bjelland Berg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gmsh

from gmsh4mrst._geometry import (
    get_perpendicular, get_extruded_points, get_midpoint,
    line_bends_towards_right, distance, calculate_number_of_points,
)

def create_transfinite_cc_box(
        start_1,
        start_2,
        start_line,
        end_1,
        end_2,
        cc_loops: list,
        cc_line_surfaces: list,
        parallel_line_points: int,
        cc_lines: list,
        perp_nodes: int = 2
    ) -> int:
    """Create a transfinite box from the supplied points
    
    Add the curve loop to cc_loops, and the surface to cc_line_surfaces.

    Return end_line
    """
    # Create the parallel lines of the box surrounding the CC line
    parallel_line_1 = gmsh.model.geo.add_line(start_2, end_1)
    parallel_line_2 = gmsh.model.geo.add_line(end_2, start_1)
    # Create the end line
    end_line = gmsh.model.geo.add_line(end_1, end_2)

    cc_lines.extend([parallel_line_1, parallel_line_2, end_line])

    # Save the curve loop, for use when creating the base surface
    curve_loop = gmsh.model.geo.add_curve_loop([
        start_line,
        parallel_line_1,
        end_line,
        parallel_line_2
    ])
    cc_loops.append(curve_loop)

    # And save the actual surface created
    surface = gmsh.model.geo.add_plane_surface([curve_loop])
    cc_line_surfaces.append(surface)

    # Create a transfinite surface out of the surface we made from
    # the box surrounding the CC line
    # End line always has 2 transfinite points, to get nice single-width cells
    gmsh.model.geo.mesh.set_transfinite_curve(end_line, perp_nodes)
    gmsh.model.geo.mesh.set_transfinite_curve(parallel_line_1, parallel_line_points)
    gmsh.model.geo.mesh.set_transfinite_curve(parallel_line_2, parallel_line_points)
    gmsh.model.geo.mesh.set_transfinite_surface(surface)
    gmsh.model.geo.mesh.set_recombine(2, surface)

    return end_line

def create_threshold_field(
        point_list: list,
        curve_list: list,
        sampling: int,
        min_size: float,
        max_size: float,
        min_distance: float,
        max_distance: float) -> int:
    """Create a Threshold field, using a Distance field as input

    Args:
        point_list (list): List of point IDs to calculate threshold for
        curve_list (list): List of curve IDs to calculate threshold for
        sampling (int): Number of sample points along the curves
        min_size (float): Minimum size of cells (close to points/curves)
        max_size (float): Maximum size of cells (far from points/curves)
        min_distance (float): Distance where ramp-up should start
        max_distance (float): Distance where ramp-up is finished

    Returns:
        int: ID of Threshold field
    """
    
    # The Distance field returns the distance to (sampling) points on each
    # fracture
    distance = gmsh.model.mesh.field.add("Distance")

    if point_list is not None:
        gmsh.model.mesh.field.setNumbers(distance, "PointsList", point_list)
    if curve_list is not None:
        gmsh.model.mesh.field.setNumbers(distance, "CurvesList", curve_list)
        gmsh.model.mesh.field.setNumber(distance, "Sampling", sampling)

    # The Treshold field uses the value from the Distance field to define a
    # change in element size depending on the computed distances
    threshold = gmsh.model.mesh.field.add("Threshold")
    gmsh.model.mesh.field.setNumber(threshold, "InField", distance)
    gmsh.model.mesh.field.setNumber(threshold, "SizeMin", min_size)
    gmsh.model.mesh.field.setNumber(threshold, "SizeMax", max_size)
    gmsh.model.mesh.field.setNumber(threshold, "DistMin", min_distance)
    gmsh.model.mesh.field.setNumber(threshold, "DistMax", max_distance)

    return threshold

def create_circumference(points: 'list[int]') -> 'list[int]':
    """Create a Gmsh circumference around a list of points.

    Args:
        points (list[int]): List of point ID to encapsulate

    Returns:
        list[int]: List of Gmsh IDs of the resulting lines
    """
    output = []
    for i in range(len(points) - 1):
        output.append(gmsh.model.geo.add_line(points[i], points[i+1]))
    output.append(gmsh.model.geo.add_line(points[-1], points[0]))
    return output

def create_fracture_point(point: tuple, intersection_IDs: dict) -> int:
    intersection_ID = intersection_IDs.get(point, "DEFAULT")
    if intersection_ID == "DEFAULT":
        return gmsh.model.geo.add_point(point[0], point[1], 0)
    elif intersection_ID is None:
        gmsh_point = gmsh.model.geo.add_point(point[0], point[1], 0)
        intersection_IDs[point] = gmsh_point
        return gmsh_point
    else:
        return intersection_IDs.get(point)

def create_cell_constraint_point(point: tuple,
                                cc_point_size: float,
                                cc_loops: list,
                                cc_point_surfaces: list,
                                cc_lines: list,
                                nodes: int = 2):
    """Handle everything needed and create transfinite "box" for CC point

    Args:
        point (tuple): The CC point, with shape (x, y)
        cc_point_size (float): The diameter of the transfinite box
        cc_loops (list): A list of CC curve loops, to append to
        cc_point_surfaces (list): A list of CC point surfaces, to append to
        cc_lines (list): A list of CC lines, to append to
    """
    x, y = point[0], point[1]
    # Create corners around the CC point - up, down, left and right
    surrounding_points = [
        gmsh.model.geo.add_point(x - cc_point_size/2, y, 0),
        gmsh.model.geo.add_point(x, y + cc_point_size/2, 0),
        gmsh.model.geo.add_point(x + cc_point_size/2, y, 0),
        gmsh.model.geo.add_point(x, y - cc_point_size/2, 0)
    ]
    # Create lines surrounding the CC point, i.e. between surrounding
    # points. These make up the mesh cell around the CC point
    surrounding_lines = create_circumference(surrounding_points)
    cc_lines.extend(surrounding_lines)

    # Save the curve loop created
    cc_loops.append(
        gmsh.model.geo.add_curve_loop(surrounding_lines)
    )
    # Save the surface created
    cc_point_surfaces.append(gmsh.model.geo.add_plane_surface([cc_loops[-1]]))
    # Make each surrounding line into a transfinite curve with 2 points
    # This creates a point in each corner, leading to a single cell
    # within the surrounding lines - naturally with a face perfectly on
    # the CC point
    for sur_line in surrounding_lines:
        gmsh.model.geo.mesh.set_transfinite_curve(sur_line, nodes)
    # Make the surface a transfinite surface
    gmsh.model.geo.mesh.set_transfinite_surface(cc_point_surfaces[-1])
    # Convert the (perfectly triangle) surface into a quadrangle one.
    gmsh.model.geo.mesh.set_recombine(2, cc_point_surfaces[-1])


def create_cell_constraint_line(line: 'list[tuple]',
                                cc_parallel_size: float,
                                cc_perpendicular_size: float,
                                cc_loops: list,
                                cc_line_surfaces: list,
                                cc_lines: list,
                                perp_nodes: int = 2):
    """Handle everything needed and create transfinite "box" for CC line


    Args:
        line (list[tuple]): The CC line, in shape [(x1, y1), (x2, y2), ...]
        cc_line_size (float): The wanted length of each line segment
        cc_loops (list): A list of CC curve loops, to append to
        cc_line_surfaces (list): A list of CC line surfaces, to append to
        cc_lines (list): A list of CC lines, to append to
    """
    # We first handle the start, then all mid points, then the end
    # Compute the normal vector of the line from start- to next point
    delta_x = line[1][0] - line[0][0]
    delta_y = line[1][1] - line[0][1]
    normal_x, normal_y = get_perpendicular(delta_x, delta_y)

    # Create starting points, one along the normal vector and one
    # opposite of the normal vector
    point_1, point_2 = get_extruded_points(line[0], normal_x, normal_y, cc_perpendicular_size)
    
    # Create actual Gmsh points of the starting points
    start_1 = gmsh.model.geo.add_point(point_1[0], point_1[1], 0)
    start_2 = gmsh.model.geo.add_point(point_2[0], point_2[1], 0)
    # Make a start line for the polygon surrounding our CC line
    start_line = gmsh.model.geo.add_line(start_1, start_2)
    cc_lines.append(start_line)
    # Convert the start line into a transfinite curve. We again use 2
    # transfinite points, leading to a width of 1 cell
    gmsh.model.geo.mesh.set_transfinite_curve(start_line, perp_nodes)

    # Handle all midpoints
    for i in range(1, len(line) - 1):
        # We find the normal vector by finding the midpoint of point
        # [i-1] and [1+1], then the vector from this midpoint to [i]
        midpoint = get_midpoint(line[i-1], line[i+1])
        normal_x = line[i][0] - midpoint[0]
        normal_y = line[i][1] - midpoint[1]
        # If the points are in a line, we simply use the normal vector
        # from earlier
        if normal_x == normal_y == 0:
            normal_x, normal_y = normal_x, normal_y
        # Like for the start point, we create two extrusions - one 
        # along the normal vector, and one opposite of it
        end_point_1, end_point_2 = get_extruded_points(
            line[i], normal_x, normal_y, cc_perpendicular_size
        )
        # If the line bends towards the right (creating an A-shape),
        # then we must flip the points. This ensures that the start-
        # and end points make a nice rectangle (compared to the twisted
        # rectangle we'd get otherwise)
        if line_bends_towards_right(line[i-1], line[i], line[i+1]):
            end_point_1, end_point_2 = end_point_2, end_point_1
        
        # Create actual Gmsh points of the extrusion points
        end_1 = gmsh.model.geo.add_point(end_point_1[0], end_point_1[1], 0)
        end_2 = gmsh.model.geo.add_point(end_point_2[0], end_point_2[1], 0)
        
        # Compute the transfinite points of the parallel lines
        # We use the line from start_2 -> end_1 as our measuring stick, as
        # the difference between the parallel lines is likely very small
        line_length = distance(line[i-1], line[i])
        parallel_line_points = calculate_number_of_points(
            line_length, cc_parallel_size
        )

        end_line = create_transfinite_cc_box(
            start_1, start_2, start_line,
            end_1, end_2,
            cc_loops, cc_line_surfaces, parallel_line_points, cc_lines,
            perp_nodes
        )

        # Convert the previous end line to a new start line
        # Note the flip (end_2 -> start_1 and vice versa)
        # This ensures we get neat rectangles for each line segment
        start_1, start_2 = end_2, end_1
        start_line = -end_line
    
    # Handle the end point
    # Same as for the start point - compute the normal vector
    delta_x = line[-1][0] - line[-2][0]
    delta_y = line[-1][1] - line[-2][1]
    normal_x, normal_y = get_perpendicular(delta_x, delta_y)
    # Use the normal vector to get extrusion points
    end_point_2, end_point_1 = get_extruded_points(line[-1], normal_x, normal_y, cc_perpendicular_size)

    # Create actual Gmsh points from the extrusion points
    end_1 = gmsh.model.geo.add_point(end_point_1[0], end_point_1[1], 0)
    end_2 = gmsh.model.geo.add_point(end_point_2[0], end_point_2[1], 0)

    # Compute the transfinite points of the parallel lines
    # We use the line from start_2 -> end_1 as our measuring stick, as
    # the difference between the parallel lines is likely very small
    line_length = distance(line[-1], line[-2])
    parallel_line_points = calculate_number_of_points(
        line_length, cc_parallel_size
    )

    # Create a transfinite surface out of the created CC box
    create_transfinite_cc_box(
        start_1, start_2, start_line,
        end_1, end_2,
        cc_loops, cc_line_surfaces, parallel_line_points,
        cc_lines, perp_nodes
    )


"""
    Helper files for internal use - Argument formatting.
    Copyright (C) 2022  Andreas Bjelland Berg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from array import array
from typing import Iterable, Union

from gmsh4mrst._assertions import (
    assert_column_in_dict, assert_columns_have_same_length
)

def format_constraints(constraints) -> 'list[list[tuple[float, float]]]':
    """Check that constraints has one of the legal forms, and format it

    Args:
        constraints (list[tuple]
                     list[list[tuple]]
                    | dict[str, float]
                    | dict[str, Iterable]
                    | dict[str, dict[str, float]]
                    | dict[str, dict[str, Iterable]]
            ): constraints that should be formatted

    Returns:
        list: The constraints in the form
            [
                [(x11, y11), (x12, y12), ...],
                [(x21, y21), (x22, y22), ...],
                ...
            ]
    """
    if isinstance(constraints, dict):
        if len(constraints) == 0:
            # constraints is an empty list -> No fractures to handle,
            # simply create a single mesh within size
            output = []
        elif isinstance(list(constraints.values())[0], array):
            # If data is sent from MATLAB, then constraints is a dict
            # If there is only one line, then we accept it as a 1D dict, i.e.
            # one can write
            #   constraints.x = [x1 ...];
            #   constraints.y = [y1 ...];
            # Then constraints is a dict in the shape {x: array, y: array}
            output = [_check_array_dict_and_return_line_list(constraints)]

        elif isinstance(list(constraints.values())[0], dict):
            # If data is sent from MATLAB, then constraints is a dict of dicts
            # If there are more than one line, then it must be a 2D dict, i.e.
            # one can write
            #   constraints.line1.x = [x11 ...];
            #   constraints.line1.y = [y11 ...];
            #   constraints.line2.x = [x21 ...];
            #   constraints.line2.y = [y21 ...];
            # Each element in constraints.values is a dict of {x: array, y: array}
            output = []
            for constraint in constraints.values():
                output.append(
                    _check_array_dict_and_return_line_list(constraint)
                )
    elif isinstance(constraints, Iterable):
        if len(constraints) == 0:
            # constraints is an empty list -> No fractures to handle,
            # simply create a single mesh within size
            output = []
        elif len(constraints) == 2 and isinstance(constraints[0], float):
            # constraints has form [x, y], likely one point
            output = [(constraints[0], constraints[1])]
        elif isinstance(constraints[0], tuple):
            # constraints is an iterable of tuples or a single point
            # -> constraints is a single list of points
            # -> check that every point contains x- and y-value
            output = []
            for constraint in constraints:
                if not isinstance(constraint, Iterable) or len(constraint) < 2:
                    raise ValueError(
                        f"Constraint {constraint} must be in the form (x, y)"
                    )
                output.append((constraint[0], constraint[1]))
        elif isinstance(constraints[0], Iterable):
            # constraints is a list lines
            # -> check that each point has x- and y-value
            output = []
            for line in constraints:
                output_line = []
                for point in line:
                    if not isinstance(point, Iterable) or len(point) < 2:
                        raise ValueError(
                            f"Constraint {point} must be in the form (x, y)"
                        )
                    output_line.append((point[0], point[1]))
                output.append(output_line)
    else:
        raise ValueError(
            f"Received unknown constraint {constraints}"
        )
    return output


def format_shape(shape) -> 'list[tuple[float, float]]':
    """Check that shape has one of the legal forms, and format it

    Args:
        shape (Iterable[float]
               | Iterable[Iterable[float]]
               | dict[str, float]
            ): Shape to format

    Raises:
        ValueError: Raised if shape has illegal type or shape

    Returns:
        The shape in the form [(x1, y1), (x2, y2), ...]
    """
    if isinstance(shape, dict):
        # Shape is a dict, likely from MATLAB
        assert_column_in_dict(shape, "x")
        assert_column_in_dict(shape, "y")
        assert_columns_have_same_length(shape, "x", "y")
        output = [
            (x, y) for x, y in zip(shape.get("x"), shape.get("y"))
        ]
    elif isinstance(shape, Iterable):
        if len(shape) < 2:
            raise ValueError(
                f"Shape must have length >= 2. Current length: {len(shape)}"
            )
        elif len(shape) == 2:
            # Shape is the size of the domain, starting at (0, 0)
            output = [
                (0, 0), (0, shape[1]), (shape[0], shape[1]), (shape[0], 0)
            ]
        else:
            # output is a list of points making up the domain
            # We check that each point consists of two actual points
            output = []
            for point in shape:
                if not isinstance(point, Iterable) or len(point) < 2:
                    raise ValueError(
                        f"Point {point} must be in the form (x, y)"
                    )
                output.append((point[0], point[1]))
    else:
        raise ValueError(
            f"Shape {shape} (type {type(shape)}) is not supported for argument `shape`!"
        )
    return output


def format_meshing_algorithm(mesh_algorithm: Union[str, int]) -> int:
    """Check that mesh_algorithm has legal value, and format it
    
    According to 
    https://gitlab.onelab.info/gmsh/gmsh/-/blob/master/tutorials/python/t10.py
    Frontal-Delaunay (6) usually leads to the highest quality meshes, but
    Delaunay (5) handles complex mesh sizes better - especially size fields
    with large element size gradients. We therefore default to Delaunay.
    For quad-shaped grids, "Frontal-Delaunay for Quads" (8) may be beneficial.

    Args:
        mesh_algorithm (str | int): Input to format

    Raises:
        ValueError: Raise if mesh_algorithm has illegal type or value

    Returns:
        int: Gmsh-ID of the meshing algorithm
    """
    mapping = {
        "meshadapt": 1,
        "automatic": 2,
        "delaunay": 5,
        "frontal": 6,
        "bamg": 7,
        "delquad": 8
    }
    if mesh_algorithm is None:
        # Use Delaunay as default
        return 5
    elif isinstance(mesh_algorithm, int):
        # Check if mesh_algorithm is a legal value. If so, use it
        if mesh_algorithm not in mapping.values():
            raise ValueError(
                "mesh_algorithm must be a legal value. Current value: "
                + f"{mesh_algorithm}. Legal values: {mapping}"
            )
        return mesh_algorithm
    elif isinstance(mesh_algorithm, str):
        if mesh_algorithm.lower() not in mapping.keys():
            raise ValueError(
                "mesh_algorithm must be a legal value. Current value: "
                + f"{mesh_algorithm}. Legal values: {mapping}"
            )
        return mapping.get(mesh_algorithm.lower())
    else:
        raise ValueError(
            f"Unsupported type for mesh_algorithm: {type(mesh_algorithm)}. "
            + "mesh_algorithm must be an int or a string."
        )


def format_recombination_algorithm(algorithm: Union[str, int]) -> int:
    """Check that recombination_algorithm has legal value, and format it
    
    Args:
        algorithm (str | int): Input to format

    Raises:
        ValueError: Raise if algorithm has illegal type or value

    Returns:
        int: Gmsh-ID of the recombination algorithm
    """
    mapping = {
        "simple": 0,
        "blossom": 1,
        "simplefull": 2,
        "blossomfull": 3
    }
    if algorithm is None:
        # Default is None
        return None
    elif isinstance(algorithm, int):
        if algorithm not in mapping.values():
            raise ValueError(
                "recombination_algorithm must be a legal value. Current value: "
                + f"{algorithm}. Legal values: {mapping}"
            )
        return algorithm
    elif isinstance(algorithm, str):
        if algorithm.lower() not in mapping.keys():
            raise ValueError(
                "recombination_algorithm must be a legal value. Current value: "
                + f"{algorithm}. Legal values: {mapping}"
            )
        return mapping.get(algorithm.lower())
    else:
        raise ValueError(
            f"Unsupported type for recombination_algorithm: {type(algorithm)}. "
            + "recombination_algorithm must be an int or a string."
        )


def format_constraint_factor(factor: Union[float, Iterable],
                             constraint_length: int) -> list:
    if isinstance(factor, float):
        return [factor] * constraint_length
    elif isinstance(factor, Iterable) and len(factor) == constraint_length:
        return factor
    elif isinstance(factor, Iterable):
        raise ValueError(
            f"Factor {factor} has wrong length. Expected length: "
            + f"{constraint_length}. Received: {len(factor)}"
        )
    else:
        raise ValueError(
            f"Factor ({factor}) of unexpected type {type(factor)} received. "
            + f"Expected type float or Iterable with length: {constraint_length}"
        )
    


def _check_array_dict_and_return_line_list(face_constraints: dict):
    # Check that face_constraints contains column x and y
    assert_column_in_dict(face_constraints, "x")
    assert_column_in_dict(face_constraints, "y")
    
    # Handle actual content
    if isinstance(face_constraints.get("x"), float):
        # If the 'line' is a single point, then x and y are floats,
        # and must be handled explicitly
        return [(face_constraints.get("x"), face_constraints.get("y"))]
    else:
        # Assume 'line' is a list/tuple/Iterable of floats
        assert_columns_have_same_length(face_constraints, "x", "y")
        return list(zip(face_constraints.get("x"), face_constraints.get("y")))